package dht;
import java.math.BigInteger;
import java.nio.charset.Charset;
import java.nio.file.FileSystems;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CopyOnWriteArrayList;
import static java.nio.file.StandardCopyOption.*;

import network.*;

import java.io.File;
import java.io.IOException;
import java.lang.*;
import java.net.*;


public class DHTNetwork implements Runnable{
	
	private String ip;
	private int port;
	private int id;
	private int storeNo= 3;
	private int tabSize = 9;
	private DHTNode[] successorList;
	private DHTNode[] predecessorList;
	private FingerTable fTable;
	private ClientNetwork network;
	private List<String> fileNames;
	private boolean transferring = false;
	private Map<String, File> upDirTable; //Keeps track of all the directories of files that we 
	//send upload requests for.
	private long dhtReplyTime = 5000;
	private long dhtCheckTime = 30000;
	private boolean dhtConfirmed = false;
	
	//Constructor for the DHT system.
	public DHTNetwork()
	{}
	
	public void initNetwork(ClientNetwork network)
	{
		this.network = network;
	}
	public void initDHT(String ip, int port, boolean firstNode)
	{
		this.ip = ip;
		this.port = port;
		id = hash(ip + port);
		fTable = new FingerTable(id, ip, port, tabSize);
		successorList = new DHTNode[storeNo];
		predecessorList = new DHTNode[storeNo];
		fileNames = new CopyOnWriteArrayList<String>();
		upDirTable = new Hashtable<String, File>();
		new Thread(this).start();
		if(firstNode)
		{
			List<String> contents = new ArrayList<String>();
			network.notifyPresentationServer("PRES_DHT_ADD", contents);
		}
	}
	
	public int getId()
	{
		return id;
	}
	
	public int getTabSize()
	{
		return tabSize;
	}
	
	//Hash function.
	public int hash(String input)
	{
		MessageDigest hash;
		byte[] digestBytes;
		BigInteger hashVal = null;
		int id;
		int idSpace;
		
		try
		{
			//Hash the input string.
			hash = MessageDigest.getInstance("SHA-1");
			hash.reset();
			hash.update(input.getBytes("utf8"));
			digestBytes = hash.digest();
			//Get the numerical value of the digest.
			hashVal = new BigInteger(digestBytes);
		}
		catch(Exception e)
		{}
		
		//Get possible values in id space.
		idSpace = (int)Math.pow(2, tabSize-1);
		//Use mod operator to get something in
		//between 0 and idSpace.
		id = hashVal.intValue()%idSpace;
		id = Math.abs(id);
		return id;
	}
	
	public DHTNode query(int itemId)
	{
		DHTNode searchNode = null;
		//Case where the current node is responsible for this
		//item id.
		if(predecessorList[0] == null)
		{
			return searchNode;
		}
		if((itemId > predecessorList[0].getId()))
		{
			if(itemId <= id)
			{
				return searchNode;
			}
			if(id < predecessorList[0].getId())
			{
				return searchNode;
			}
		}
		if((itemId < predecessorList[0].getId()))
		{
			if((id < predecessorList[0].getId())&&(id >= itemId))
			{
				return searchNode;
			}
		}

		searchNode = fTable.findSuccessor(itemId);
			
		if(searchNode.getId() == id)
		{
			searchNode = null;
		}
		
		return searchNode;
	}
	
	public void printDetails()
	{
		System.out.println("ID: " + id);
		System.out.println("Address: " + ip);
		System.out.println("Port: " + port);
		
		if(successorList[0] != null)
		{
			System.out.println("Successor 1: " + successorList[0].getId());
		}
		else
		{
			System.out.println("Successor 1: Not Set");
		}
		
		if(successorList[1] != null)
		{
			System.out.println("Successor 2: " + successorList[1].getId());
		}
		else
		{
			System.out.println("Successor 2: Not Set");
		}
		
		if(successorList[2] != null)
		{
			System.out.println("Successor 3: " + successorList[2].getId());
		}
		else
		{
			System.out.println("Successor 3: Not Set");
		}
		
		if(predecessorList[0] != null)
		{
			System.out.println("Predecessor 1: " + predecessorList[0].getId());
		}
		else
		{
			System.out.println("Predecessor 1: Not Set");
		}
		
		if(predecessorList[1] != null)
		{
			System.out.println("Predecessor 2: " + predecessorList[1].getId());
		}
		else
		{
			System.out.println("Predecessor 2: Not Set");
		}
		
		if(predecessorList[2] != null)
		{
			System.out.println("Predecessor 3: " + predecessorList[2].getId());
		}
		else
		{
			System.out.println("Predecessor 3: Not Set");
		}
		fTable.printTable();
		printFileNames();
	}
	
	public void addNode(String ip, int port)
	{
		int newId;
		
		newId = hash(ip + port);
		fTable.addNode(newId, ip, port);
		
		if(!attemptSetSuccessor(ip, port, 0))
		{
			if(!attemptSetSuccessor(ip, port, 1))
			{
				attemptSetSuccessor(ip, port, 2);
			}
		}
		
		if(!attemptSetPredecessor(ip, port, 0))
		{
			if(!attemptSetPredecessor(ip, port, 1))
			{
				attemptSetPredecessor(ip, port, 2);
			}
		}
		
		System.out.println("Adding " + newId);
	}
	
	public void joinNetwork(String destAddress, int destPort)
	{
		//Send a join request to the provided node. That node will then
		//locate the correct successor. 
		List<String> packetData = new ArrayList<String>();
		int length;
		
		packetData.add(ip);
		packetData.add(Integer.toString(port));
		length = network.addData("DHT_JOIN", packetData,
				destAddress, destPort);
	}
	
	public void forwardJoinReq(List<String> packetContents)
	{
		int destId;
		DHTNode fwdNode;
		List<String> packetData = new ArrayList<String>();
		int length;
		String destAddress;
		int destPort;
		DHTNode orgNode;
		
		//Retrieve new node information from packet.
		orgNode = removeDHTInfo(packetContents);
		destAddress = orgNode.getAddress();
		destPort = orgNode.getPort();
		
		//Get the closest successor in the finger table.
		destId = hash(destAddress + destPort);
		fwdNode = query(destId);
		
		//Forward the request to them if it's not us and if it is us,
		//then reply to the request and tell them what we know.
		//This means it's us.
		if(fwdNode == null)
		{
			//Push all the predecessors back and have this new node as the immediate
			//predecessor.
			if(destId == id)
			{
				return;
			}
			returnDetails(destAddress, destPort);
			/*
			fTable.addNode(destId, destAddress, destPort);
			predecessorList[2] = predecessorList[1];
			predecessorList[1] = predecessorList[0];
			predecessorList[0] = new DHTNode(destId, destAddress, destPort);
			*/
			addNode(destAddress, destPort);
			System.out.println("Served join request.");
			printDetails();
		}
		//This means it's not us. Reconstruct this packet and send it
		//to the node we think it belongs to.
		else
		{
			packetData.add(destAddress);
			packetData.add(Integer.toString(destPort));
			length = network.addData("DHT_JOIN", packetData,
					fwdNode.getAddress(), fwdNode.getPort());
		}
	}
	
	//Communicate an addition to the predecessor.
	public void notifySelfAddition()
	{
		try
		{
			commAddition(predecessorList[0], network.getClientAddress(), network.getClientPort());
		}
		catch(UnknownHostException uhException)
		{
			System.out.println("Could not send addition packet.");
		}
	}
	
	//Sends an addition notice to a node for the input address
	//and port combination.
	public void commAddition(DHTNode node, String address, int port)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		String destAddress;
		int destPort;
	
		if(node == null)
		{
			return;
		}
			
		destAddress = node.getAddress();
		destPort = node.getPort();
			
		packetData.add(address);
		packetData.add(Integer.toString(port));
			
		length = network.addData("DHT_ADD", packetData,
					destAddress, destPort);	
	}
	
	public void forwardAddition(List<String> packetContents)
	{
		String newAddress;
		int newPort;
		DHTNode newNode;
		
		newNode = removeDHTInfo(packetContents);
		
		if(newNode != null)
		{
			//Stop forwarding if it has come back all the way to you.
			if(newNode.getId() == id)
			{
				return;
			}
			newAddress = newNode.getAddress();
			newPort = newNode.getPort();
			//Add the node for ourselves.
			addNode(newAddress, newPort);
		
			//Forward the addition notice backwards in the chord circle.
			commAddition(predecessorList[0], newAddress, newPort);
		}
	}
	
	//This method deals with the compilation of all the successor, predecessor
	//and finger table information when a new predecessor is added
	//to the network.
	public void returnDetails(String destAddress, int destPort)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		List<DHTNode> list;
		
		list = fTable.getNodes();
		
		//Add all of the unique nodes in the finger table. We'll know
		//when finger table information begins because there's
		//always going to be ONLY 5 entries before this.
		while(list.size() > 0)
		{
			addDHTInfo(list.remove(0), packetData);
		}
		//The last successor doesn't need to be sent because
		//this node will be one of them.
		addDHTInfo(successorList[1], packetData);
		addDHTInfo(successorList[0], packetData);
		addDHTInfo(predecessorList[2], packetData);
		addDHTInfo(predecessorList[1], packetData);
		addDHTInfo(predecessorList[0], packetData);
		
		length = network.addData("DHT_SETUP", packetData,
				destAddress, destPort);
	}
	
	public void receiveDetails(List<String> packetData, String senderAddress, int senderPort)
	{
		int senderId;
		
		//Make the first successor from the address and port that the packet
		//originated from.
		senderId = hash(senderAddress + senderPort);
		
		//Proceed to recreate the nodes on this side.
		System.out.println("Adding predecessors.");
		predecessorList[0] = removeDHTInfo(packetData);
		predecessorList[1] = removeDHTInfo(packetData);
		predecessorList[2] = removeDHTInfo(packetData);
		System.out.println("Adding successors.");
		successorList[0] = new DHTNode(senderId, senderAddress, senderPort);
		successorList[1] = removeDHTInfo(packetData);
		successorList[2] = removeDHTInfo(packetData);
		//Of course our immediate successor needs to be added also.
		addNode(senderAddress, senderPort);
		
		System.out.println("Adding finger table successors.");
		while(packetData.size() > 0)
		{
			//After this point, it's only finger table information.
			fTable.addNode(removeDHTInfo(packetData));
		}
		notifySelfAddition();
	}
	
	//Adds node address and port to packet data to prepare for transmission.
	public void addDHTInfo(DHTNode tempNode, List<String> packetData)
	{
		if(tempNode == null)
		{
			packetData.add("null");
			packetData.add("null");
		}
		else
		{
			packetData.add(tempNode.getAddress());
			packetData.add(Integer.toString(tempNode.getPort()));
		}
	}
	
	//Removes node address and port from packet data when received.
	public DHTNode removeDHTInfo(List<String> packetData)
	{
		DHTNode returnNode;
		int returnId;
		String returnAddress;
		int returnPort;
		String strPort;
		
		//Remove the two last fields for port and address.
		strPort = packetData.remove(packetData.size()-1);
		
		if(strPort.equalsIgnoreCase("null"))
		{
			packetData.remove(packetData.size()-1);
			return null;
		}
		
		returnPort = Integer.parseInt(strPort);
		returnAddress = packetData.remove(packetData.size()-1);
		returnId = hash(returnAddress + returnPort);
		
		//Turn into new node and return that object.
		returnNode = new DHTNode(returnId, returnAddress, returnPort); 
		return returnNode;
	}
	
	//Method to set one of the successors to a new node.
	public boolean attemptSetSuccessor(String newIp, int newPort, int index)
	{
		int newId;
		DHTNode tempNode;
		
		newId = hash(newIp + newPort);
		tempNode = new DHTNode(newId, newIp, newPort);
		
		if(newId == id)
		{
			return true;
		}
		
		if(successorList[index] == null)
		{
			setSuccessor(tempNode, index);
			return true;
		}
		else if(newId == successorList[index].getId())
		{
			return true;
		}
		else if(newId > id)
		{
			if(successorList[index].getId() < id)
			{
				setSuccessor(tempNode, index);
				return true;
			}
			if(successorList[index].getId() > newId)
			{
				setSuccessor(tempNode, index);
				return true;
			}
		}
		else
		{
			if(successorList[index].getId() > id)
			{
				return false;
			}
			if(successorList[index].getId() > newId)
			{
				setSuccessor(tempNode, index);
				return true;
			}
		}
		return false;
	}
	
	public void setSuccessor(DHTNode newNode, int index)
	{
		int counter;
		
		//Shift back every successor after the one we're 
		//inserting.
		for(counter=(storeNo-1); counter>index; counter--)
		{
			successorList[counter] = successorList[counter-1];
		}
		
		successorList[index] = newNode;
	}
	
	//Method to set one of the predecessors to a new node.
	public boolean attemptSetPredecessor(String newIp, int newPort, int index)
	{
		int newId;
		DHTNode tempNode;
		
		newId = hash(newIp + newPort);
		tempNode = new DHTNode(newId, newIp, newPort);
		
		if(newId == id)
		{
			return true;
		}
		
		if(predecessorList[index] == null)
		{
			setPredecessor(tempNode, index);
			return true;
		}
		else if(newId == predecessorList[index].getId())
		{
			return true;
		}
		else if(newId > id)
		{
			if(predecessorList[index].getId() < id)
			{
				return false;
			}
			if(predecessorList[index].getId() < newId)
			{
				setPredecessor(tempNode, index);
				return true;
			}
		}
		else
		{
			if(predecessorList[index].getId() > id)
			{
				setPredecessor(tempNode, index);
				return true;
			}
			if(predecessorList[index].getId() < newId)
			{
				setPredecessor(tempNode, index);
				return true;
			}
		}
		return false;
	}
	
	public void setPredecessor(DHTNode newNode, int index)
	{
		int counter;
		
		//Shift back every successor after the one we're 
		//inserting.
		for(counter=(storeNo-1); counter>index; counter--)
		{
			predecessorList[counter] = predecessorList[counter-1];
		}
		
		predecessorList[index] = newNode;
		
		if(index == 0)
		{
			checkFiles();
		}
	}
	
	public void uploadCheck(String fileName, String reqAddress, int reqPort, File selectedUpload)
	{
		DHTNode checkNode;
		int fileId;
		int filePort;
		
		fileId = hash(fileName);
		checkNode = query(fileId);
		
		if(checkNode == null)
		{
			List<String> packetData = new ArrayList<String>();
			packetData.add(fileName);
			System.out.println("Received upload req for ID " + fileId);
			network.setLastUploadId(id);
			transferFile(selectedUpload);
			network.notifyPresentationServer("PRES_FILE", packetData);
			return;
		}
		
		filePort = sendFile(selectedUpload);
		forwardFileUpRequest(fileName, reqAddress, reqPort, filePort);
	}
	
	public boolean forwardFileUpRequest(String fileName, String reqAddress, int reqPort, int filePort)
	{
		DHTNode fwdNode;
		List<String> packetData = new ArrayList<String>();
		int length;
		int fileId;
		
		fileId = hash(fileName);
		
		fwdNode = query(fileId);
		
		//Forward the request to them if it's not us and if it is us,
		//then reply to the request and tell them what we know.
		//This means it's us.
		if(fwdNode == null)
		{
			System.out.println("Received upload req for ID " + fileId);
			addFileName(fileName);
			return true;
		}
		//This means it's not us. Reconstruct this packet and send it
		//to the node we think it belongs to.
		else
		{
			packetData.add(fileName);
			packetData.add(reqAddress);
			packetData.add(Integer.toString(reqPort));
			packetData.add(Integer.toString(filePort));
			length = network.addData("DHT_UP", packetData,
					fwdNode.getAddress(), fwdNode.getPort());
			network.setLastUploadId(fwdNode.getId());
			return false;
		}
	}
	
	public void returnFileUpConfirmation(String fileName, String reqAddress, int reqPort, int filePort)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		
		packetData.add(fileName);
		length = network.addData("DHT_UP_CONFIRM", packetData,
				reqAddress, reqPort);
		receiveFile(fileName, reqAddress, filePort);
	}
	
	public void receiveUpRequest(List<String> packetContents)
	{
		String fileName;
		String reqAddress;
		int reqPort;
		int filePort;
		
		fileName = packetContents.remove(0);
		reqAddress = packetContents.remove(0);
		reqPort = Integer.parseInt(packetContents.remove(0));
		filePort = Integer.parseInt(packetContents.remove(0));
		
		if(forwardFileUpRequest(fileName, reqAddress, reqPort, filePort))
		{
			returnFileUpConfirmation(fileName, reqAddress, reqPort, filePort);
		}
	}
	
	public void receiveFile(String fileName, String address, int port)
	{
		Runnable r = new FileReceiver(fileName, address, port, this);
		new Thread(r).start();
	}
	
	public int sendFile(File newFile)
	{
		FileSender r = new FileSender(newFile, this);
		int portNo = r.getPort();
		new Thread(r).start();
		return portNo;
	}
	
	public boolean forwardFileDownRequest(String fileName, String reqAddress, int reqPort, boolean initRequest)
	{
		DHTNode fwdNode;
		List<String> packetData = new ArrayList<String>();
		int length;
		int fileId;
		
		fileId = hash(fileName);
		
		fwdNode = query(fileId);
		
		//Forward the request to them if it's not us and if it is us,
		//then reply to the request and tell them what we know.
		//This means it's us.
		if(fwdNode == null)
		{
			System.out.println("Received download req for ID " + fileId);
			if(initRequest)
			{
				network.setLastDownloadId(id);
			}
			return true;
		}
		//This means it's not us. Reconstruct this packet and send it
		//to the node we think it belongs to.
		else
		{
			packetData.add(fileName);
			packetData.add(reqAddress);
			packetData.add(Integer.toString(reqPort));
			length = network.addData("DHT_DOWN", packetData,
					fwdNode.getAddress(), fwdNode.getPort());
			if(initRequest)
			{
				network.setLastDownloadId(fwdNode.getId());
			}
			return false;
		}
	}
	
	public void returnFileDownConfirmation(String fileName, String reqAddress, int reqPort)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		int filePort;
		File targetFile;
		
		targetFile = new File(fileName);
		
		filePort = sendFile(targetFile);
		packetData.add(fileName);
		packetData.add(Integer.toString(filePort));
		length = network.addData("DHT_DOWN_CONFIRM", packetData,
				reqAddress, reqPort);
	}
	
	public void returnFileTransfer(String fileName, String reqAddress, int reqPort)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		int filePort;
		File targetFile;
		
		targetFile = new File(fileName);
		filePort = sendFile(targetFile);
		packetData.add(fileName);
		packetData.add(Integer.toString(filePort));
		length = network.addData("DHT_TRANSFER", packetData,
				reqAddress, reqPort);
	}
	
	public void receiveDownRequest(List<String> packetContents)
	{
		String fileName;
		String reqAddress;
		int reqPort;
		
		fileName = packetContents.remove(0);
		reqAddress = packetContents.remove(0);
		reqPort = Integer.parseInt(packetContents.remove(0));
		
		if(forwardFileDownRequest(fileName, reqAddress, reqPort, false))
		{
			returnFileDownConfirmation(fileName, reqAddress, reqPort);
		}
	}
	
	public void notifySelfRemoval()
	{
		int fileNo;
		transferring = true;
		List<String> dupeList = new ArrayList<String>();
		if(successorList[0] != null)
		{
			try
			{
				commRemoval(predecessorList[0], network.getClientAddress(), network.getClientPort(),
						predecessorList[0], predecessorList[1], predecessorList[2], successorList[0],
						successorList[1], successorList[2]);
			}
			catch(UnknownHostException uhException)
			{
				System.out.println("Could not send removal packet.");
			}
		}
		
		int counter;
		
		//Send a file transfer requests for each file.
		if(predecessorList[0] != null)
		{
			fileNo = fileNames.size();
			dupeList.addAll(fileNames);
			for(counter=0;counter<fileNo;counter++)
			{
				returnFileTransfer(dupeList.get(counter),
						successorList[0].getAddress(), 
						successorList[0].getPort());
			}
			
			
			//DON'T close until we've finished shifting all our stuff.
			while(fileNames.size() > 0)
			{
			}
			transferring = false;
		}
	}
	
	//Sends an addition notice to a node for the input address
	//and port combination.
	public void commRemoval(DHTNode node, String address, int port, 
			DHTNode pred0, DHTNode pred1, DHTNode pred2, DHTNode succ0,
			DHTNode succ1, DHTNode succ2)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		String destAddress;
		int destPort;
		int delId;
		
		delId = hash(address + port);
			
		destAddress = node.getAddress();
		destPort = node.getPort();
			
		packetData.add(address);
		packetData.add(Integer.toString(port));
		addDHTInfo(pred0, packetData);
		addDHTInfo(pred1, packetData);
		addDHTInfo(pred2, packetData);
		addDHTInfo(succ0, packetData);
		addDHTInfo(succ1, packetData);
		addDHTInfo(succ2, packetData);
			
		length = network.addData("DHT_REMOVAL", packetData,
					destAddress, destPort);	
	}
	
	public void forwardRemoval(List<String> packetContents)
	{
		String delAddress;
		int delPort;
		DHTNode delNode;
		DHTNode pred0;
		DHTNode pred1;
		DHTNode pred2;
		DHTNode succ0;
		DHTNode succ1;
		DHTNode succ2;
		
		succ2 = removeDHTInfo(packetContents);
		succ1 = removeDHTInfo(packetContents);
		succ0 = removeDHTInfo(packetContents);
		pred2 = removeDHTInfo(packetContents);
		pred1 = removeDHTInfo(packetContents);
		pred0 = removeDHTInfo(packetContents);
		delNode = removeDHTInfo(packetContents);

		delAddress = delNode.getAddress();
		delPort = delNode.getPort();

		//Forward the addition notice backwards in the chord circle.
		//Only if we've reached the other side of the chord circle
		//can we stop.
		if(predecessorList[0].getId() != delNode.getId())
		{
			commRemoval(predecessorList[0], delAddress, delPort, pred0, pred1,
					pred2, succ0, succ1, succ2);
		}
		
		//Clear the node.
		clearNode(delNode);
		printDetails();
		//Add all of the nodes in case one of them will patch the hole the
		//removed node left.
		if(succ0 != null)
		{
			addNode(succ0.getAddress(), succ0.getPort());
		}
		
		if(succ1 != null)
		{
			addNode(succ1.getAddress(), succ1.getPort());
		}
		
		if(succ2 != null)
		{
			addNode(succ2.getAddress(), succ2.getPort());
		}
		
		if(pred0 != null)
		{
			addNode(pred0.getAddress(), pred0.getPort());
		}
		
		if(pred1 != null)
		{
			addNode(pred1.getAddress(), pred1.getPort());
		}
		
		if(pred2 != null)
		{
			addNode(pred2.getAddress(), pred2.getPort());
		}
		printDetails();
	}
	
	public void clearNode(DHTNode oldNode)
	{
		//Firstly replace the deleted node with ourself (the least optimal
		//node before trying to add the rest). Typically, the successor 0
		//should take it's place but we don't need to dictate that in code.
		//Our add function should take appropriate action.
		fTable.replaceNode(oldNode, new DHTNode(id, ip, port));
		
		attemptRemovePredecessor(oldNode, 0);
		attemptRemovePredecessor(oldNode, 1);
		attemptRemovePredecessor(oldNode, 2);
		attemptRemoveSuccessor(oldNode, 0);
		attemptRemoveSuccessor(oldNode, 1);
		attemptRemoveSuccessor(oldNode, 2);
	}
	
	public void attemptRemovePredecessor(DHTNode oldNode, int index)
	{
		int counter;
		
		if(predecessorList[index] == null)
		{
			return;
		}
		
		if(oldNode.getId() != predecessorList[index].getId())
		{
			return;
		}
		
		//Push everything forwards.
		for(counter=index; counter<(storeNo-1); counter++)
		{
			predecessorList[counter] = predecessorList[counter+1];
		}
		
		predecessorList[storeNo-1] = null;
	}
	
	public void attemptRemoveSuccessor(DHTNode oldNode, int index)
	{
		int counter;
		
		if(successorList[index] == null)
		{
			return;
		}
		
		if(oldNode.getId() != successorList[index].getId())
		{
			return;
		}
		
		//Shift back every successor after the one we're 
		//inserting.
		for(counter=index; counter<(storeNo-1); counter++)
		{
			successorList[counter] = successorList[counter+1];
		}
		
		successorList[storeNo-1] = null;
	}
	
	public void addFileName(String newName)
	{
		fileNames.add(newName);
		network.refreshDHTGUI();
	}
	
	public void removeFileName(String newName)
	{
		int counter;
		
		for(counter=0; counter<fileNames.size(); counter++)
		{
			if(fileNames.get(counter).equalsIgnoreCase(newName))
			{
				fileNames.remove(counter);
				network.refreshDHTGUI();
				return;
			}
		}
	}
	
	public void printFileNames()
	{
		int counter;
		int fileId;
		
		for(counter=0; counter<fileNames.size(); counter++)
		{
			fileId = hash(fileNames.get(counter));
			System.out.println(fileNames.get(counter) + " (" + fileId + ")");
		}
	}
	
	public void checkFiles()
	{
		String predAddress;
		int predPort;
		int fileId;
		int counter;
		int fileNo;
		List<String> dupeList = new ArrayList<String>();
		
		predAddress = predecessorList[0].getAddress();
		predPort = predecessorList[0].getPort();
		fileNo = fileNames.size();
		dupeList.addAll(fileNames);
		
		for(counter=0;counter<fileNo;counter++)
		{
			fileId = hash(dupeList.get(counter));
			if(query(fileId) != null)
			{
				returnFileTransfer(dupeList.get(counter),
						predAddress, predPort);
				removeFileName(dupeList.get(counter));
			}
		}
	}
	
	public boolean isTransferring()
	{
		return transferring;
	}
	
	public void addUpEntry(String fileName, File newFile)
	{
		upDirTable.put(fileName, newFile);
	}
	
	//Method is used to send check packets to the predecessor to make sure
	//it's alive.
	public void run()
	{
		String predAddress;
		int predPort;
		List<String> packetData;
		DHTNode delNode;
		
		while(true)
		{
			dhtConfirmed = false;
			System.out.println("Loop begins...");
			if(predecessorList[0] != null)
			{
				predAddress = predecessorList[0].getAddress();
				predPort = predecessorList[0].getPort();
				
				System.out.println("Checking predecessor.");
				packetData = new ArrayList<String>();
				network.addData("DHT_PRED_CHECK", packetData, 
						predAddress, predPort);
				try
				{
					Thread.sleep(dhtReplyTime);
				}
				catch(InterruptedException iException)
				{
					System.out.println("DHT predecessor check interrupted.");
					continue;
				}
				
				if(dhtConfirmed == false)
				{
					System.out.println("Predecessor failed to respond to life check.");
					delNode = new DHTNode(predecessorList[0].getId(),
							predecessorList[0].getAddress(),
							predecessorList[0].getPort());
					
					if(successorList[0]!= null)
					{
						if(successorList[0].getId() == delNode.getId())
						{
							commDeath(successorList[0], delNode.getAddress(),
									delNode.getPort(), predecessorList[1],
									predecessorList[2], null, successorList[0],
									successorList[1], successorList[2]);
						}
					}
					clearNode(delNode);
					printDetails();
				}
			}
			try
			{
				Thread.sleep(dhtCheckTime);
			}
			catch(InterruptedException iException)
			{
				System.out.println("DHT predecessor check interrupted.");
			}
		}
	}
	
	public void setDHTConfirmation(boolean dhtConfirmed)
	{
		this.dhtConfirmed = dhtConfirmed;
	}
	
	public void commDeath(DHTNode node, String address, int port, 
			DHTNode pred0, DHTNode pred1, DHTNode pred2, DHTNode succ0,
			DHTNode succ1, DHTNode succ2)
	{
		List<String> packetData = new ArrayList<String>();
		int length;
		String destAddress;
		int destPort;
		int delId;
		
		delId = hash(address + port);
			
		destAddress = node.getAddress();
		destPort = node.getPort();
			
		packetData.add(address);
		packetData.add(Integer.toString(port));
		addDHTInfo(pred0, packetData);
		addDHTInfo(pred1, packetData);
		addDHTInfo(pred2, packetData);
		addDHTInfo(succ0, packetData);
		addDHTInfo(succ1, packetData);
		addDHTInfo(succ2, packetData);
			
		length = network.addData("DHT_DEATH", packetData,
					destAddress, destPort);	
	}
	
	public void forwardDeath(List<String> packetContents)
	{
		String delAddress;
		int delPort;
		DHTNode delNode;
		DHTNode pred0;
		DHTNode pred1;
		DHTNode pred2;
		DHTNode succ0;
		DHTNode succ1;
		DHTNode succ2;
		
		succ2 = removeDHTInfo(packetContents);
		succ1 = removeDHTInfo(packetContents);
		succ0 = removeDHTInfo(packetContents);
		pred2 = removeDHTInfo(packetContents);
		pred1 = removeDHTInfo(packetContents);
		pred0 = removeDHTInfo(packetContents);
		delNode = removeDHTInfo(packetContents);

		delAddress = delNode.getAddress();
		delPort = delNode.getPort();

		//Forward the addition notice backwards in the chord circle.
		//Only if we've reached the other side of the chord circle
		//can we stop.
		if(successorList[0].getId() != delNode.getId())
		{
			commDeath(successorList[0], delAddress, delPort, pred0, pred1,
					pred2, succ0, succ1, succ2);
		}
		else
		{
			sendFix(successorList[1], predecessorList[2]);
		}
		
		//Clear the node.
		clearNode(delNode);
		printDetails();
		//Add all of the nodes in case one of them will patch the hole the
		//removed node left.
		if(succ0 != null)
		{
			addNode(succ0.getAddress(), succ0.getPort());
		}
		
		if(succ1 != null)
		{
			addNode(succ1.getAddress(), succ1.getPort());
		}
		
		if(succ2 != null)
		{
			addNode(succ2.getAddress(), succ2.getPort());
		}
		
		if(pred0 != null)
		{
			addNode(pred0.getAddress(), pred0.getPort());
		}
		
		if(pred1 != null)
		{
			addNode(pred1.getAddress(), pred1.getPort());
		}
		
		if(pred2 != null)
		{
			addNode(pred2.getAddress(), pred2.getPort());
		}
		printDetails();
	}
	
	//When a recipient node needs to know about a new 
	//successor/predecessor to fill it's lists or finger table,
	//we use this.
	public void sendFix(DHTNode recipient, DHTNode addition)
	{
		List<String> packetData = new ArrayList<String>();
		
		addDHTInfo(addition, packetData);
		network.addData("DHT_FIX", packetData,
				recipient.getAddress(), recipient.getPort());	
	}
	
	public boolean fileNameExists(String fileName)
	{
		int counter;
		
		for(counter=0;counter<fileNames.size();counter++)
		{
			if(fileName.equalsIgnoreCase(fileNames.get(counter)))
			{
				return true;
			}
		}
		return false;
	}
	
	public String[][] getTableInfo()
	{
		return fTable.getTableInfo();
	}
	
	public String[][] getListInfo()
	{
		String[][] stringList = new String[6][3];
		
		addListInfo(stringList, predecessorList[0], 0);
		addListInfo(stringList, predecessorList[1], 1);
		addListInfo(stringList, predecessorList[2], 2);
		addListInfo(stringList, successorList[0], 3);
		addListInfo(stringList, successorList[1], 4);
		addListInfo(stringList, successorList[2], 5);
		
		return stringList;
	}
	
	public void addListInfo(String[][] stringList, DHTNode addNode, int index)
	{
		if(addNode != null)
		{
			stringList[index][0] = Integer.toString(addNode.getId());
			stringList[index][1] = addNode.getAddress();
			stringList[index][2] = Integer.toString(addNode.getPort());
		}
		else
		{
			stringList[index][0] = "Not set";
			stringList[index][1] = "Not set";
			stringList[index][2] = "Not set";
		}
	}
	
	public String[][] getFileNames()
	{
		String[][] fileList;
		fileList = new String[2][fileNames.size()];
		int counter;
		
		for(counter=0;counter<fileNames.size();counter++)
		{
			fileList[0][counter] = fileNames.get(counter);
			fileList[1][counter] = Integer.toString(hash(fileNames.get(counter)));
		}
		
		return fileList;
	}
	
	public void transferFile(File transFile)
	{
		Path currentPath = transFile.toPath();
		File newFile = new File(transFile.getName());
		Path newPath = newFile.toPath();
		
		try {
			Files.copy(currentPath, newPath, REPLACE_EXISTING);
		} catch (IOException e) {
			
		}
		
		addFileName(transFile.getName());
	}
}
